# Copyright (C) 2006-2012, 2015, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the urlutils wrapper."""

import os
import sys

from .. import osutils, urlutils
from ..errors import PathNotChild
from . import TestCase, TestCaseInTempDir, TestSkipped, features


class TestUrlToPath(TestCase):

    def test_basename(self):
        # breezy.urlutils.basename
        # Test breezy.urlutils.split()
        basename = urlutils.basename
        if sys.platform == 'win32':
            self.assertRaises(urlutils.InvalidURL, basename,
                              'file:///path/to/foo')
            self.assertEqual('foo', basename('file:///C|/foo'))
            self.assertEqual('foo', basename('file:///C:/foo'))
            self.assertEqual('', basename('file:///C:/'))
        else:
            self.assertEqual('foo', basename('file:///foo'))
            self.assertEqual('', basename('file:///'))

        self.assertEqual('foo', basename('http://host/path/to/foo'))
        self.assertEqual('foo', basename('http://host/path/to/foo/'))
        self.assertEqual(
            '', basename('http://host/path/to/foo/',
                         exclude_trailing_slash=False))
        self.assertEqual('path', basename('http://host/path'))
        self.assertEqual('', basename('http://host/'))
        self.assertEqual('', basename('http://host'))
        self.assertEqual('path', basename('http:///nohost/path'))

        self.assertEqual('path', basename(
            'random+scheme://user:pass@ahost:port/path'))
        self.assertEqual('path', basename(
            'random+scheme://user:pass@ahost:port/path/'))
        self.assertEqual('', basename('random+scheme://user:pass@ahost:port/'))

        # relative paths
        self.assertEqual('foo', basename('path/to/foo'))
        self.assertEqual('foo', basename('path/to/foo/'))
        self.assertEqual('', basename('path/to/foo/',
                                      exclude_trailing_slash=False))
        self.assertEqual('foo', basename('path/../foo'))
        self.assertEqual('foo', basename('../path/foo'))

    def test_normalize_url_files(self):
        # Test that local paths are properly normalized
        normalize_url = urlutils.normalize_url

        def norm_file(expected, path):
            url = normalize_url(path)
            self.assertStartsWith(url, 'file:///')
            if sys.platform == 'win32':
                url = url[len('file:///C:'):]
            else:
                url = url[len('file://'):]

            self.assertEndsWith(url, expected)

        norm_file('path/to/foo', 'path/to/foo')
        norm_file('/path/to/foo', '/path/to/foo')
        norm_file('path/to/foo', '../path/to/foo')

        # Local paths are assumed to *not* be escaped at all
        try:
            'uni/\xb5'.encode(osutils.get_user_encoding())
        except UnicodeError:
            # locale cannot handle unicode
            pass
        else:
            norm_file('uni/%C2%B5', 'uni/\xb5')

        norm_file('uni/%25C2%25B5', 'uni/%C2%B5')
        norm_file('uni/%20b', 'uni/ b')
        # All the crazy characters get escaped in local paths => file:/// urls
        # The ' ' character must not be at the end, because on win32
        # it gets stripped off by ntpath.abspath
        norm_file('%27%20%3B/%3F%3A%40%26%3D%2B%24%2C%23', "' ;/?:@&=+$,#")

    def test_normalize_url_hybrid(self):
        # Anything with a scheme:// should be treated as a hybrid url
        # which changes what characters get escaped.
        normalize_url = urlutils.normalize_url

        eq = self.assertEqual
        eq('file:///foo/', normalize_url('file:///foo/'))
        eq('file:///foo/%20', normalize_url('file:///foo/ '))
        eq('file:///foo/%20', normalize_url('file:///foo/%20'))
        # Don't escape reserved characters
        eq('file:///ab_c.d-e/%f:?g&h=i+j;k,L#M$',
            normalize_url('file:///ab_c.d-e/%f:?g&h=i+j;k,L#M$'))
        eq('http://ab_c.d-e/%f:?g&h=i+j;k,L#M$',
            normalize_url('http://ab_c.d-e/%f:?g&h=i+j;k,L#M$'))

        # Escape unicode characters, but not already escaped chars
        eq('http://host/ab/%C2%B5/%C2%B5',
            normalize_url('http://host/ab/%C2%B5/\xb5'))

        # Unescape characters that don't need to be escaped
        eq('http://host/~bob%2525-._',
           normalize_url('http://host/%7Ebob%2525%2D%2E%5F'))
        eq('http://host/~bob%2525-._',
           normalize_url('http://host/%7Ebob%2525%2D%2E%5F'))

    def test_url_scheme_re(self):
        # Test paths that may be URLs
        def test_one(url, scheme_and_path):
            """Assert that _url_scheme_re correctly matches

            :param scheme_and_path: The (scheme, path) that should be matched
                can be None, to indicate it should not match
            """
            m = urlutils._url_scheme_re.match(url)
            if scheme_and_path is None:
                self.assertEqual(None, m)
            else:
                self.assertEqual(scheme_and_path[0], m.group('scheme'))
                self.assertEqual(scheme_and_path[1], m.group('path'))

        # Local paths
        test_one('/path', None)
        test_one('C:/path', None)
        test_one('../path/to/foo', None)
        test_one('../path/to/fo\xe5', None)

        # Real URLS
        test_one('http://host/path/', ('http', 'host/path/'))
        test_one('sftp://host/path/to/foo', ('sftp', 'host/path/to/foo'))
        test_one('file:///usr/bin', ('file', '/usr/bin'))
        test_one('file:///C:/Windows', ('file', '/C:/Windows'))
        test_one('file:///C|/Windows', ('file', '/C|/Windows'))
        test_one('readonly+sftp://host/path/\xe5',
                 ('readonly+sftp', 'host/path/\xe5'))

        # Weird stuff
        # Can't have slashes or colons in the scheme
        test_one('/path/to/://foo', None)
        test_one('scheme:stuff://foo', ('scheme', 'stuff://foo'))
        # Must have more than one character for scheme
        test_one('C://foo', None)
        test_one('ab://foo', ('ab', 'foo'))

    def test_dirname(self):
        # Test breezy.urlutils.dirname()
        dirname = urlutils.dirname
        if sys.platform == 'win32':
            self.assertRaises(urlutils.InvalidURL, dirname,
                              'file:///path/to/foo')
            self.assertEqual('file:///C|/', dirname('file:///C|/foo'))
            self.assertEqual('file:///C|/', dirname('file:///C|/'))
        else:
            self.assertEqual('file:///', dirname('file:///foo'))
            self.assertEqual('file:///', dirname('file:///'))

        self.assertEqual('http://host/path/to',
                         dirname('http://host/path/to/foo'))
        self.assertEqual('http://host/path/to',
                         dirname('http://host/path/to/foo/'))
        self.assertEqual('http://host/path/to/foo',
                         dirname('http://host/path/to/foo/',
                                 exclude_trailing_slash=False))
        self.assertEqual('http://host/', dirname('http://host/path'))
        self.assertEqual('http://host/', dirname('http://host/'))
        self.assertEqual('http://host', dirname('http://host'))
        self.assertEqual('http:///nohost', dirname('http:///nohost/path'))

        self.assertEqual('random+scheme://user:pass@ahost:port/',
                         dirname('random+scheme://user:pass@ahost:port/path'))
        self.assertEqual('random+scheme://user:pass@ahost:port/',
                         dirname('random+scheme://user:pass@ahost:port/path/'))
        self.assertEqual('random+scheme://user:pass@ahost:port/',
                         dirname('random+scheme://user:pass@ahost:port/'))

        # relative paths
        self.assertEqual('path/to', dirname('path/to/foo'))
        self.assertEqual('path/to', dirname('path/to/foo/'))
        self.assertEqual('path/to/foo',
                         dirname('path/to/foo/', exclude_trailing_slash=False))
        self.assertEqual('path/..', dirname('path/../foo'))
        self.assertEqual('../path', dirname('../path/foo'))

    def test_is_url(self):
        self.assertTrue(urlutils.is_url('http://foo/bar'))
        self.assertTrue(urlutils.is_url('bzr+ssh://foo/bar'))
        self.assertTrue(urlutils.is_url('lp:foo/bar'))
        self.assertTrue(urlutils.is_url('file:///foo/bar'))
        self.assertFalse(urlutils.is_url(''))
        self.assertFalse(urlutils.is_url('foo'))
        self.assertFalse(urlutils.is_url('foo/bar'))
        self.assertFalse(urlutils.is_url('/foo'))
        self.assertFalse(urlutils.is_url('/foo/bar'))
        self.assertFalse(urlutils.is_url('C:/'))
        self.assertFalse(urlutils.is_url('C:/foo'))
        self.assertFalse(urlutils.is_url('C:/foo/bar'))

    def test_join(self):
        def test(expected, *args):
            joined = urlutils.join(*args)
            self.assertEqual(expected, joined)

        # Test relative path joining
        test('foo', 'foo')  # relative fragment with nothing is preserved.
        test('foo/bar', 'foo', 'bar')
        test('http://foo/bar', 'http://foo', 'bar')
        test('http://foo/bar', 'http://foo', '.', 'bar')
        test('http://foo/baz', 'http://foo', 'bar', '../baz')
        test('http://foo/bar/baz', 'http://foo', 'bar/baz')
        test('http://foo/baz', 'http://foo', 'bar/../baz')
        test('http://foo/baz', 'http://foo/bar/', '../baz')
        test('lp:foo/bar', 'lp:foo', 'bar')
        test('lp:foo/bar/baz', 'lp:foo', 'bar/baz')

        # Absolute paths
        test('http://foo', 'http://foo')  # abs url with nothing is preserved.
        test('http://bar', 'http://foo', 'http://bar')
        test('sftp://bzr/foo', 'http://foo', 'bar', 'sftp://bzr/foo')
        test('file:///bar', 'foo', 'file:///bar')
        test('http://bar/', 'http://foo', 'http://bar/')
        test('http://bar/a', 'http://foo', 'http://bar/a')
        test('http://bar/a/', 'http://foo', 'http://bar/a/')
        test('lp:bar', 'http://foo', 'lp:bar')
        test('lp:bar', 'lp:foo', 'lp:bar')
        test('file:///stuff', 'lp:foo', 'file:///stuff')

        # From a base path
        test('file:///foo', 'file:///', 'foo')
        test('file:///bar/foo', 'file:///bar/', 'foo')
        test('http://host/foo', 'http://host/', 'foo')
        test('http://host/', 'http://host', '')

        # Invalid joinings
        # Cannot go above root
        # Implicitly at root:
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.join,
                          'http://foo', '../baz')
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.join,
                          'http://foo', '/..')
        # Joining from a path explicitly under the root.
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.join,
                          'http://foo/a', '../../b')

    def test_joinpath(self):
        def test(expected, *args):
            joined = urlutils.joinpath(*args)
            self.assertEqual(expected, joined)

        # Test a single element
        test('foo', 'foo')

        # Test relative path joining
        test('foo/bar', 'foo', 'bar')
        test('foo/bar', 'foo', '.', 'bar')
        test('foo/baz', 'foo', 'bar', '../baz')
        test('foo/bar/baz', 'foo', 'bar/baz')
        test('foo/baz', 'foo', 'bar/../baz')

        # Test joining to an absolute path
        test('/foo', '/foo')
        test('/foo', '/foo', '.')
        test('/foo/bar', '/foo', 'bar')
        test('/', '/foo', '..')

        # Test joining with an absolute path
        test('/bar', 'foo', '/bar')

        # Test joining to a path with a trailing slash
        test('foo/bar', 'foo/', 'bar')

        # Invalid joinings
        # Cannot go above root
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.joinpath, '/',
                          '../baz')
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.joinpath, '/',
                          '..')
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.joinpath, '/',
                          '/..')

    def test_join_segment_parameters_raw(self):
        join_segment_parameters_raw = urlutils.join_segment_parameters_raw
        self.assertEqual("/somedir/path",
                         join_segment_parameters_raw("/somedir/path"))
        self.assertEqual("/somedir/path,rawdata",
                         join_segment_parameters_raw(
                             "/somedir/path", "rawdata"))
        self.assertRaises(urlutils.InvalidURLJoin,
                          join_segment_parameters_raw, "/somedir/path",
                          "rawdata1,rawdata2,rawdata3")
        self.assertEqual("/somedir/path,bla,bar",
                         join_segment_parameters_raw(
                             "/somedir/path", "bla", "bar"))
        self.assertEqual(
            "/somedir,exist=some/path,bla,bar",
            join_segment_parameters_raw("/somedir,exist=some/path",
                                        "bla", "bar"))
        self.assertRaises(TypeError, join_segment_parameters_raw,
                          "/somepath", 42)

    def test_join_segment_parameters(self):
        join_segment_parameters = urlutils.join_segment_parameters
        self.assertEqual("/somedir/path",
                         join_segment_parameters("/somedir/path", {}))
        self.assertEqual(
            "/somedir/path,key1=val1",
            join_segment_parameters("/somedir/path", {"key1": "val1"}))
        self.assertRaises(urlutils.InvalidURLJoin,
                          join_segment_parameters, "/somedir/path",
                          {"branch": "brr,brr,brr"})
        self.assertRaises(
            urlutils.InvalidURLJoin,
            join_segment_parameters, "/somedir/path", {"key1=val1": "val2"})
        self.assertEqual("/somedir/path,key1=val1,key2=val2",
                         join_segment_parameters("/somedir/path", {
                             "key1": "val1", "key2": "val2"}))
        self.assertEqual("/somedir/path,key1=val1,key2=val2",
                         join_segment_parameters("/somedir/path,key1=val1", {
                             "key2": "val2"}))
        self.assertEqual("/somedir/path,key1=val2",
                         join_segment_parameters("/somedir/path,key1=val1", {
                             "key1": "val2"}))
        self.assertEqual("/somedir,exist=some/path,key1=val1",
                         join_segment_parameters("/somedir,exist=some/path",
                                                 {"key1": "val1"}))
        self.assertEqual(
            "/,key1=val1,key2=val2",
            join_segment_parameters("/,key1=val1", {"key2": "val2"}))
        self.assertRaises(TypeError,
                          join_segment_parameters, "/,key1=val1", {"foo": 42})

    def test_function_type(self):
        if sys.platform == 'win32':
            self.assertEqual(urlutils._win32_local_path_to_url,
                             urlutils.local_path_to_url)
            self.assertEqual(urlutils._win32_local_path_from_url,
                             urlutils.local_path_from_url)
        else:
            self.assertEqual(urlutils._posix_local_path_to_url,
                             urlutils.local_path_to_url)
            self.assertEqual(urlutils._posix_local_path_from_url,
                             urlutils.local_path_from_url)

    def test_posix_local_path_to_url(self):
        to_url = urlutils._posix_local_path_to_url
        self.assertEqual('file:///path/to/foo',
                         to_url('/path/to/foo'))

        self.assertEqual('file:///path/to/foo%2Cbar',
                         to_url('/path/to/foo,bar'))

        try:
            result = to_url('/path/to/r\xe4ksm\xf6rg\xe5s')
        except UnicodeError:
            raise TestSkipped("local encoding cannot handle unicode")

        self.assertEqual('file:///path/to/r%C3%A4ksm%C3%B6rg%C3%A5s', result)
        self.assertTrue(isinstance(result, str))

    def test_posix_local_path_from_url(self):
        from_url = urlutils._posix_local_path_from_url
        self.assertEqual('/path/to/foo',
                         from_url('file:///path/to/foo'))
        self.assertEqual('/path/to/foo',
                         from_url('file:///path/to/foo,branch=foo'))
        self.assertEqual('/path/to/r\xe4ksm\xf6rg\xe5s',
                         from_url('file:///path/to/r%C3%A4ksm%C3%B6rg%C3%A5s'))
        self.assertEqual('/path/to/r\xe4ksm\xf6rg\xe5s',
                         from_url('file:///path/to/r%c3%a4ksm%c3%b6rg%c3%a5s'))
        self.assertEqual(
            '/path/to/r\xe4ksm\xf6rg\xe5s',
            from_url('file://localhost/path/to/r%c3%a4ksm%c3%b6rg%c3%a5s'))

        self.assertRaises(urlutils.InvalidURL, from_url, '/path/to/foo')
        self.assertRaises(
            urlutils.InvalidURL, from_url,
            'file://remotehost/path/to/r%c3%a4ksm%c3%b6rg%c3%a5s')

    def test_win32_local_path_to_url(self):
        to_url = urlutils._win32_local_path_to_url
        self.assertEqual('file:///C:/path/to/foo',
                         to_url('C:/path/to/foo'))
        # BOGUS: on win32, ntpath.abspath will strip trailing
        #       whitespace, so this will always fail
        #       Though under linux, it fakes abspath support
        #       and thus will succeed
        # self.assertEqual('file:///C:/path/to/foo%20',
        #     to_url('C:/path/to/foo '))
        self.assertEqual('file:///C:/path/to/f%20oo',
                         to_url('C:/path/to/f oo'))

        self.assertEqual('file:///', to_url('/'))

        self.assertEqual('file:///C:/path/to/foo%2Cbar',
                         to_url('C:/path/to/foo,bar'))
        try:
            result = to_url('d:/path/to/r\xe4ksm\xf6rg\xe5s')
        except UnicodeError:
            raise TestSkipped("local encoding cannot handle unicode")

        self.assertEqual(
            'file:///D:/path/to/r%C3%A4ksm%C3%B6rg%C3%A5s', result)
        self.assertIsInstance(result, str)

    def test_win32_unc_path_to_url(self):
        self.requireFeature(features.win32_feature)
        to_url = urlutils._win32_local_path_to_url
        self.assertEqual('file://HOST/path',
                         to_url(r'\\HOST\path'))
        self.assertEqual('file://HOST/path',
                         to_url('//HOST/path'))

        try:
            result = to_url('//HOST/path/to/r\xe4ksm\xf6rg\xe5s')
        except UnicodeError:
            raise TestSkipped("local encoding cannot handle unicode")

        self.assertEqual(
            'file://HOST/path/to/r%C3%A4ksm%C3%B6rg%C3%A5s', result)
        self.assertFalse(isinstance(result, str))

    def test_win32_local_path_from_url(self):
        from_url = urlutils._win32_local_path_from_url
        self.assertEqual('C:/path/to/foo',
                         from_url('file:///C|/path/to/foo'))
        self.assertEqual(
            'D:/path/to/r\xe4ksm\xf6rg\xe5s',
            from_url('file:///d|/path/to/r%C3%A4ksm%C3%B6rg%C3%A5s'))
        self.assertEqual(
            'D:/path/to/r\xe4ksm\xf6rg\xe5s',
            from_url('file:///d:/path/to/r%c3%a4ksm%c3%b6rg%c3%a5s'))
        self.assertEqual('/', from_url('file:///'))
        self.assertEqual('C:/path/to/foo',
                         from_url('file:///C|/path/to/foo,branch=foo'))

        self.assertRaises(urlutils.InvalidURL, from_url, 'file:///C:')
        self.assertRaises(urlutils.InvalidURL, from_url, 'file:///c')
        self.assertRaises(urlutils.InvalidURL, from_url, '/path/to/foo')
        # Not a valid _win32 url, no drive letter
        self.assertRaises(urlutils.InvalidURL, from_url, 'file:///path/to/foo')

    def test_win32_unc_path_from_url(self):
        from_url = urlutils._win32_local_path_from_url
        self.assertEqual('//HOST/path', from_url('file://HOST/path'))
        self.assertEqual('//HOST/path',
                         from_url('file://HOST/path,branch=foo'))
        # despite IE allows 2, 4, 5 and 6 slashes in URL to another machine
        # we want to use only 2 slashes
        # Firefox understand only 5 slashes in URL, but it's ugly
        self.assertRaises(urlutils.InvalidURL, from_url, 'file:////HOST/path')
        self.assertRaises(urlutils.InvalidURL, from_url, 'file://///HOST/path')
        self.assertRaises(urlutils.InvalidURL, from_url,
                          'file://////HOST/path')
        # check for file://C:/ instead of file:///C:/
        self.assertRaises(urlutils.InvalidURL, from_url, 'file://C:/path')

    def test_win32_extract_drive_letter(self):
        extract = urlutils._win32_extract_drive_letter
        self.assertEqual(('file:///C:', '/foo'), extract('file://', '/C:/foo'))
        self.assertEqual(('file:///d|', '/path'),
                         extract('file://', '/d|/path'))
        self.assertRaises(urlutils.InvalidURL, extract, 'file://', '/path')
        # Root drives without slash treated as invalid, see bug #841322
        self.assertEqual(('file:///C:', '/'), extract('file://', '/C:/'))
        self.assertRaises(urlutils.InvalidURL, extract, 'file://', '/C:')
        # Invalid without drive separator or following forward slash
        self.assertRaises(urlutils.InvalidURL, extract, 'file://', '/C')
        self.assertRaises(urlutils.InvalidURL, extract, 'file://', '/C:ool')

    def test_split(self):
        # Test breezy.urlutils.split()
        split = urlutils.split
        if sys.platform == 'win32':
            self.assertRaises(urlutils.InvalidURL, split,
                              'file:///path/to/foo')
            self.assertEqual(('file:///C|/', 'foo'), split('file:///C|/foo'))
            self.assertEqual(('file:///C:/', ''), split('file:///C:/'))
        else:
            self.assertEqual(('file:///', 'foo'), split('file:///foo'))
            self.assertEqual(('file:///', ''), split('file:///'))

        self.assertEqual(('http://host/path/to', 'foo'),
                         split('http://host/path/to/foo'))
        self.assertEqual(('http://host/path/to', 'foo'),
                         split('http://host/path/to/foo/'))
        self.assertEqual(
            ('http://host/path/to/foo', ''),
            split('http://host/path/to/foo/', exclude_trailing_slash=False))
        self.assertEqual(('http://host/', 'path'), split('http://host/path'))
        self.assertEqual(('http://host/', ''), split('http://host/'))
        self.assertEqual(('http://host', ''), split('http://host'))
        self.assertEqual(('http:///nohost', 'path'),
                         split('http:///nohost/path'))

        self.assertEqual(('random+scheme://user:pass@ahost:port/', 'path'),
                         split('random+scheme://user:pass@ahost:port/path'))
        self.assertEqual(('random+scheme://user:pass@ahost:port/', 'path'),
                         split('random+scheme://user:pass@ahost:port/path/'))
        self.assertEqual(('random+scheme://user:pass@ahost:port/', ''),
                         split('random+scheme://user:pass@ahost:port/'))

        # relative paths
        self.assertEqual(('path/to', 'foo'), split('path/to/foo'))
        self.assertEqual(('path/to', 'foo'), split('path/to/foo/'))
        self.assertEqual(('path/to/foo', ''),
                         split('path/to/foo/', exclude_trailing_slash=False))
        self.assertEqual(('path/..', 'foo'), split('path/../foo'))
        self.assertEqual(('../path', 'foo'), split('../path/foo'))

    def test_strip_segment_parameters(self):
        strip_segment_parameters = urlutils.strip_segment_parameters
        # Check relative references with absolute paths
        self.assertEqual("/some/path",
                         strip_segment_parameters("/some/path"))
        self.assertEqual("/some/path",
                         strip_segment_parameters("/some/path,tip"))
        self.assertEqual("/some,dir/path",
                         strip_segment_parameters("/some,dir/path,tip"))
        self.assertEqual(
            "/somedir/path",
            strip_segment_parameters("/somedir/path,heads%2Ftip"))
        self.assertEqual(
            "/somedir/path",
            strip_segment_parameters("/somedir/path,heads%2Ftip,bar"))
        # Check relative references with relative paths
        self.assertEqual("", strip_segment_parameters(",key1=val1"))
        self.assertEqual("foo/", strip_segment_parameters("foo/,key1=val1"))
        self.assertEqual("foo", strip_segment_parameters("foo,key1=val1"))
        self.assertEqual(
            "foo/base,la=bla/other/elements",
            strip_segment_parameters("foo/base,la=bla/other/elements"))
        self.assertEqual(
            "foo/base,la=bla/other/elements",
            strip_segment_parameters("foo/base,la=bla/other/elements,a=b"))
        # TODO: Check full URLs as well as relative references

    def test_split_segment_parameters_raw(self):
        split_segment_parameters_raw = urlutils.split_segment_parameters_raw
        # Check relative references with absolute paths
        self.assertEqual(("/some/path", []),
                         split_segment_parameters_raw("/some/path"))
        self.assertEqual(("/some/path", ["tip"]),
                         split_segment_parameters_raw("/some/path,tip"))
        self.assertEqual(("/some,dir/path", ["tip"]),
                         split_segment_parameters_raw("/some,dir/path,tip"))
        self.assertEqual(
            ("/somedir/path", ["heads%2Ftip"]),
            split_segment_parameters_raw("/somedir/path,heads%2Ftip"))
        self.assertEqual(
            ("/somedir/path", ["heads%2Ftip", "bar"]),
            split_segment_parameters_raw("/somedir/path,heads%2Ftip,bar"))
        # Check relative references with relative paths
        self.assertEqual(("", ["key1=val1"]),
                         split_segment_parameters_raw(",key1=val1"))
        self.assertEqual(("foo/", ["key1=val1"]),
                         split_segment_parameters_raw("foo/,key1=val1"))
        self.assertEqual(("foo", ["key1=val1"]),
                         split_segment_parameters_raw("foo,key1=val1"))
        self.assertEqual(
            ("foo/base,la=bla/other/elements", []),
            split_segment_parameters_raw("foo/base,la=bla/other/elements"))
        self.assertEqual(
            ("foo/base,la=bla/other/elements", ["a=b"]),
            split_segment_parameters_raw("foo/base,la=bla/other/elements,a=b"))
        # TODO: Check full URLs as well as relative references

    def test_split_segment_parameters(self):
        split_segment_parameters = urlutils.split_segment_parameters
        # Check relative references with absolute paths
        self.assertEqual(("/some/path", {}),
                         split_segment_parameters("/some/path"))
        self.assertEqual(("/some/path", {"branch": "tip"}),
                         split_segment_parameters("/some/path,branch=tip"))
        self.assertEqual(("/some,dir/path", {"branch": "tip"}),
                         split_segment_parameters("/some,dir/path,branch=tip"))
        self.assertEqual(
            ("/somedir/path", {"ref": "heads%2Ftip"}),
            split_segment_parameters("/somedir/path,ref=heads%2Ftip"))
        self.assertEqual(("/somedir/path",
                          {"ref": "heads%2Ftip", "key1": "val1"}),
                         split_segment_parameters(
            "/somedir/path,ref=heads%2Ftip,key1=val1"))
        self.assertEqual(
            ("/somedir/path", {"ref": "heads%2F=tip"}),
            split_segment_parameters("/somedir/path,ref=heads%2F=tip"))
        # Check relative references with relative paths
        self.assertEqual(("", {"key1": "val1"}),
                         split_segment_parameters(",key1=val1"))
        self.assertEqual(("foo/", {"key1": "val1"}),
                         split_segment_parameters("foo/,key1=val1"))
        self.assertEqual(
            ("foo/base,key1=val1/other/elements", {}),
            split_segment_parameters("foo/base,key1=val1/other/elements"))
        self.assertEqual(("foo/base,key1=val1/other/elements",
                          {"key2": "val2"}), split_segment_parameters(
            "foo/base,key1=val1/other/elements,key2=val2"))
        self.assertRaises(
            urlutils.InvalidURL, split_segment_parameters,
            "foo/base,key1")
        # TODO: Check full URLs as well as relative references

    def test_win32_strip_local_trailing_slash(self):
        strip = urlutils._win32_strip_local_trailing_slash
        self.assertEqual('file://', strip('file://'))
        self.assertEqual('file:///', strip('file:///'))
        self.assertEqual('file:///C', strip('file:///C'))
        self.assertEqual('file:///C:', strip('file:///C:'))
        self.assertEqual('file:///d|', strip('file:///d|'))
        self.assertEqual('file:///C:/', strip('file:///C:/'))
        self.assertEqual('file:///C:/a', strip('file:///C:/a/'))

    def test_strip_trailing_slash(self):
        sts = urlutils.strip_trailing_slash
        if sys.platform == 'win32':
            self.assertEqual('file:///C|/', sts('file:///C|/'))
            self.assertEqual('file:///C:/foo', sts('file:///C:/foo'))
            self.assertEqual('file:///C|/foo', sts('file:///C|/foo/'))
        else:
            self.assertEqual('file:///', sts('file:///'))
            self.assertEqual('file:///foo', sts('file:///foo'))
            self.assertEqual('file:///foo', sts('file:///foo/'))

        self.assertEqual('http://host/', sts('http://host/'))
        self.assertEqual('http://host/foo', sts('http://host/foo'))
        self.assertEqual('http://host/foo', sts('http://host/foo/'))

        # No need to fail just because the slash is missing
        self.assertEqual('http://host', sts('http://host'))
        # TODO: jam 20060502 Should this raise InvalidURL?
        self.assertEqual('file://', sts('file://'))

        self.assertEqual('random+scheme://user:pass@ahost:port/path',
                         sts('random+scheme://user:pass@ahost:port/path'))
        self.assertEqual('random+scheme://user:pass@ahost:port/path',
                         sts('random+scheme://user:pass@ahost:port/path/'))
        self.assertEqual('random+scheme://user:pass@ahost:port/',
                         sts('random+scheme://user:pass@ahost:port/'))

        # Make sure relative paths work too
        self.assertEqual('path/to/foo', sts('path/to/foo'))
        self.assertEqual('path/to/foo', sts('path/to/foo/'))
        self.assertEqual('../to/foo', sts('../to/foo/'))
        self.assertEqual('path/../foo', sts('path/../foo/'))

    def test_unescape_for_display_utf8(self):
        # Test that URLs are converted to nice unicode strings for display
        def test(expected, url, encoding='utf-8'):
            disp_url = urlutils.unescape_for_display(url, encoding=encoding)
            self.assertIsInstance(disp_url, str)
            self.assertEqual(expected, disp_url)

        test('http://foo', 'http://foo')
        if sys.platform == 'win32':
            test('C:/foo/path', 'file:///C|/foo/path')
            test('C:/foo/path', 'file:///C:/foo/path')
        else:
            test('/foo/path', 'file:///foo/path')

        test('http://foo/%2Fbaz', 'http://foo/%2Fbaz')
        test('http://host/r\xe4ksm\xf6rg\xe5s',
             'http://host/r%C3%A4ksm%C3%B6rg%C3%A5s')

        # Make sure special escaped characters stay escaped
        test('http://host/%3B%2F%3F%3A%40%26%3D%2B%24%2C%23',
             'http://host/%3B%2F%3F%3A%40%26%3D%2B%24%2C%23')

        # Can we handle sections that don't have utf-8 encoding?
        test('http://host/%EE%EE%EE/r\xe4ksm\xf6rg\xe5s',
             'http://host/%EE%EE%EE/r%C3%A4ksm%C3%B6rg%C3%A5s')

        # Test encoding into output that can handle some characters
        test('http://host/%EE%EE%EE/r\xe4ksm\xf6rg\xe5s',
             'http://host/%EE%EE%EE/r%C3%A4ksm%C3%B6rg%C3%A5s',
             encoding='iso-8859-1')

        # This one can be encoded into utf8
        test('http://host/\u062c\u0648\u062c\u0648',
             'http://host/%d8%ac%d9%88%d8%ac%d9%88',
             encoding='utf-8')

        # This can't be put into 8859-1 and so stays as escapes
        test('http://host/%d8%ac%d9%88%d8%ac%d9%88',
             'http://host/%d8%ac%d9%88%d8%ac%d9%88',
             encoding='iso-8859-1')

    def test_escape(self):
        self.assertEqual('%25', urlutils.escape('%'))
        self.assertEqual('%C3%A5', urlutils.escape('\xe5'))
        self.assertIsInstance(urlutils.escape('\xe5'), str)

    def test_escape_tildes(self):
        self.assertEqual('~foo', urlutils.escape('~foo'))

    def test_unescape(self):
        self.assertEqual('%', urlutils.unescape('%25'))
        self.assertEqual('\xe5', urlutils.unescape('%C3%A5'))

        self.assertEqual('\xe5', urlutils.unescape('%C3%A5'))

    def test_escape_unescape(self):
        self.assertEqual('\xe5', urlutils.unescape(urlutils.escape('\xe5')))
        self.assertEqual('%', urlutils.unescape(urlutils.escape('%')))

    def test_relative_url(self):
        def test(expected, base, other):
            result = urlutils.relative_url(base, other)
            self.assertEqual(expected, result)

        test('a', 'http://host/', 'http://host/a')
        test('http://entirely/different', 'sftp://host/branch',
             'http://entirely/different')
        test('../person/feature', 'http://host/branch/mainline',
             'http://host/branch/person/feature')
        test('..', 'http://host/branch', 'http://host/')
        test('http://host2/branch', 'http://host1/branch',
             'http://host2/branch')
        test('.', 'http://host1/branch', 'http://host1/branch')
        test('../../../branch/2b', 'file:///home/jelmer/foo/bar/2b',
             'file:///home/jelmer/branch/2b')
        test('../../branch/2b', 'sftp://host/home/jelmer/bar/2b',
             'sftp://host/home/jelmer/branch/2b')
        test('../../branch/feature/%2b', 'http://host/home/jelmer/bar/%2b',
             'http://host/home/jelmer/branch/feature/%2b')
        test('../../branch/feature/2b', 'http://host/home/jelmer/bar/2b/',
             'http://host/home/jelmer/branch/feature/2b')
        # relative_url should preserve a trailing slash
        test('../../branch/feature/2b/', 'http://host/home/jelmer/bar/2b/',
             'http://host/home/jelmer/branch/feature/2b/')
        test('../../branch/feature/2b/', 'http://host/home/jelmer/bar/2b',
             'http://host/home/jelmer/branch/feature/2b/')

        # TODO: treat http://host as http://host/
        #       relative_url is typically called from a branch.base or
        #       transport.base which always ends with a /
        # test('a', 'http://host', 'http://host/a')
        test('http://host/a', 'http://host', 'http://host/a')
        # test('.', 'http://host', 'http://host/')
        test('http://host/', 'http://host', 'http://host/')
        # test('.', 'http://host/', 'http://host')
        test('http://host', 'http://host/', 'http://host')

        # On Windows file:///C:/path/to and file:///D:/other/path
        # should not use relative url over the non-existent '/' directory.
        if sys.platform == 'win32':
            # on the same drive
            test('../../other/path',
                 'file:///C:/path/to', 'file:///C:/other/path')
            # ~next two tests is failed, i.e. urlutils.relative_url expects
            # ~to see normalized file URLs?
            # ~test('../../other/path',
            # ~    'file:///C:/path/to', 'file:///c:/other/path')
            # ~test('../../other/path',
            # ~    'file:///C:/path/to', 'file:///C|/other/path')

            # check UNC paths too
            test('../../other/path',
                 'file://HOST/base/path/to', 'file://HOST/base/other/path')
            # on different drives
            test('file:///D:/other/path',
                 'file:///C:/path/to', 'file:///D:/other/path')
            # TODO: strictly saying in UNC path //HOST/base is full analog
            # of drive letter for hard disk, and this situation is also
            # should be exception from rules. [bialix 20071221]


class TestCwdToURL(TestCaseInTempDir):
    """Test that local_path_to_url works based on the cwd"""

    def test_dot(self):
        # This test will fail if getcwd is not ascii
        os.mkdir('mytest')
        os.chdir('mytest')

        url = urlutils.local_path_to_url('.')
        self.assertEndsWith(url, '/mytest')

    def test_non_ascii(self):
        try:
            os.mkdir('dod\xe9')
        except UnicodeError:
            raise TestSkipped('cannot create unicode directory')

        os.chdir('dod\xe9')

        # On Mac OSX this directory is actually:
        #   u'/dode\u0301' => '/dode\xcc\x81
        # but we should normalize it back to
        #   u'/dod\xe9' => '/dod\xc3\xa9'
        url = urlutils.local_path_to_url('.')
        self.assertEndsWith(url, '/dod%C3%A9')


class TestDeriveToLocation(TestCase):
    """Test that the mapping of FROM_LOCATION to TO_LOCATION works."""

    def test_to_locations_derived_from_paths(self):
        derive = urlutils.derive_to_location
        self.assertEqual("bar", derive("bar"))
        self.assertEqual("bar", derive("../bar"))
        self.assertEqual("bar", derive("/foo/bar"))
        self.assertEqual("bar", derive("c:/foo/bar"))
        self.assertEqual("bar", derive("c:bar"))

    def test_to_locations_derived_from_urls(self):
        derive = urlutils.derive_to_location
        self.assertEqual("bar", derive("http://foo/bar"))
        self.assertEqual("bar", derive("bzr+ssh://foo/bar"))
        self.assertEqual("foo-bar", derive("lp:foo-bar"))


class TestRebaseURL(TestCase):
    """Test the behavior of rebase_url."""

    def test_non_relative(self):
        result = urlutils.rebase_url('file://foo', 'file://foo',
                                     'file://foo/bar')
        self.assertEqual('file://foo', result)
        result = urlutils.rebase_url('/foo', 'file://foo',
                                     'file://foo/bar')
        self.assertEqual('/foo', result)

    def test_different_ports(self):
        e = self.assertRaises(urlutils.InvalidRebaseURLs, urlutils.rebase_url,
                              'foo', 'http://bar:80', 'http://bar:81')
        self.assertEqual(str(e), "URLs differ by more than path:"
                         " 'http://bar:80' and 'http://bar:81'")

    def test_different_hosts(self):
        e = self.assertRaises(urlutils.InvalidRebaseURLs, urlutils.rebase_url,
                              'foo', 'http://bar', 'http://baz')
        self.assertEqual(str(e), "URLs differ by more than path: 'http://bar'"
                         " and 'http://baz'")

    def test_different_protocol(self):
        e = self.assertRaises(urlutils.InvalidRebaseURLs, urlutils.rebase_url,
                              'foo', 'http://bar', 'ftp://bar')
        self.assertEqual(str(e), "URLs differ by more than path: 'http://bar'"
                         " and 'ftp://bar'")

    def test_rebase_success(self):
        self.assertEqual('../bar', urlutils.rebase_url('bar', 'http://baz/',
                                                       'http://baz/qux'))
        self.assertEqual(
            'qux/bar',
            urlutils.rebase_url('bar', 'http://baz/qux', 'http://baz/'))
        self.assertEqual(
            '.', urlutils.rebase_url('foo', 'http://bar/', 'http://bar/foo/'))
        self.assertEqual(
            'qux/bar',
            urlutils.rebase_url('../bar', 'http://baz/qux/foo', 'http://baz/'))

    def test_determine_relative_path(self):
        self.assertEqual('../../baz/bar',
                         urlutils.determine_relative_path(
                             '/qux/quxx', '/baz/bar'))
        self.assertEqual('..',
                         urlutils.determine_relative_path(
                             '/bar/baz', '/bar'))
        self.assertEqual('baz',
                         urlutils.determine_relative_path(
                             '/bar', '/bar/baz'))
        self.assertEqual('.', urlutils.determine_relative_path(
                         '/bar', '/bar'))


class TestParseURL(TestCase):

    def test_parse_simple(self):
        parsed = urlutils.parse_url('http://example.com:80/one')
        self.assertEqual(('http', None, None, 'example.com', 80, '/one'),
                         parsed)

    def test_ipv6(self):
        parsed = urlutils.parse_url('http://[1:2:3::40]/one')
        self.assertEqual(('http', None, None, '1:2:3::40', None, '/one'),
                         parsed)

    def test_ipv6_port(self):
        parsed = urlutils.parse_url('http://[1:2:3::40]:80/one')
        self.assertEqual(('http', None, None, '1:2:3::40', 80, '/one'),
                         parsed)


class TestURL(TestCase):

    def test_parse_simple(self):
        parsed = urlutils.URL.from_string('http://example.com:80/one')
        self.assertEqual('http', parsed.scheme)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual('example.com', parsed.host)
        self.assertEqual(80, parsed.port)
        self.assertEqual('/one', parsed.path)

    def test_ipv6(self):
        parsed = urlutils.URL.from_string('http://[1:2:3::40]/one')
        self.assertEqual('http', parsed.scheme)
        self.assertIs(None, parsed.port)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual('1:2:3::40', parsed.host)
        self.assertEqual('/one', parsed.path)

    def test_ipv6_port(self):
        parsed = urlutils.URL.from_string('http://[1:2:3::40]:80/one')
        self.assertEqual('http', parsed.scheme)
        self.assertEqual('1:2:3::40', parsed.host)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual(80, parsed.port)
        self.assertEqual('/one', parsed.path)

    def test_quoted(self):
        parsed = urlutils.URL.from_string(
            'http://ro%62ey:h%40t@ex%41mple.com:2222/path')
        self.assertEqual(parsed.quoted_host, 'ex%41mple.com')
        self.assertEqual(parsed.host, 'exAmple.com')
        self.assertEqual(parsed.port, 2222)
        self.assertEqual(parsed.quoted_user, 'ro%62ey')
        self.assertEqual(parsed.user, 'robey')
        self.assertEqual(parsed.quoted_password, 'h%40t')
        self.assertEqual(parsed.password, 'h@t')
        self.assertEqual(parsed.path, '/path')

    def test_eq(self):
        parsed1 = urlutils.URL.from_string('http://[1:2:3::40]:80/one')
        parsed2 = urlutils.URL.from_string('http://[1:2:3::40]:80/one')
        self.assertEqual(parsed1, parsed2)
        self.assertEqual(parsed1, parsed1)
        parsed2.path = '/two'
        self.assertNotEqual(parsed1, parsed2)

    def test_repr(self):
        parsed = urlutils.URL.from_string('http://[1:2:3::40]:80/one')
        self.assertEqual(
            "<URL('http', None, None, '1:2:3::40', 80, '/one')>",
            repr(parsed))

    def test_str(self):
        parsed = urlutils.URL.from_string('http://[1:2:3::40]:80/one')
        self.assertEqual('http://[1:2:3::40]:80/one', str(parsed))

    def test__combine_paths(self):
        combine = urlutils.URL._combine_paths
        self.assertEqual('/home/sarah/project/foo',
                         combine('/home/sarah', 'project/foo'))
        self.assertEqual('/etc',
                         combine('/home/sarah', '../../etc'))
        self.assertEqual('/etc',
                         combine('/home/sarah', '../../../etc'))
        self.assertEqual('/etc',
                         combine('/home/sarah', '/etc'))

    def test_clone(self):
        url = urlutils.URL.from_string('http://[1:2:3::40]:80/one')
        url1 = url.clone("two")
        self.assertEqual("/one/two", url1.path)
        url2 = url.clone("/two")
        self.assertEqual("/two", url2.path)
        url3 = url.clone()
        self.assertIsNot(url, url3)
        self.assertEqual(url, url3)

    def test_parse_empty_port(self):
        parsed = urlutils.URL.from_string('http://example.com:/one')
        self.assertEqual('http', parsed.scheme)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual('example.com', parsed.host)
        self.assertIs(None, parsed.port)
        self.assertEqual('/one', parsed.path)


class TestFileRelpath(TestCase):

    # GZ 2011-11-18: A way to override all path handling functions to one
    #                platform or another for testing would be nice.

    def _with_posix_paths(self):
        self.overrideAttr(urlutils, "local_path_from_url",
                          urlutils._posix_local_path_from_url)
        self.overrideAttr(urlutils, "MIN_ABS_FILEURL_LENGTH", len("file:///"))
        self.overrideAttr(osutils, "normpath", osutils._posix_normpath)
        self.overrideAttr(osutils, "abspath", osutils.posixpath.abspath)
        self.overrideAttr(osutils, "normpath", osutils._posix_normpath)
        self.overrideAttr(osutils, "pathjoin", osutils.posixpath.join)
        self.overrideAttr(osutils, "split", osutils.posixpath.split)
        self.overrideAttr(osutils, "MIN_ABS_PATHLENGTH", 1)

    def _with_win32_paths(self):
        self.overrideAttr(urlutils, "local_path_from_url",
                          urlutils._win32_local_path_from_url)
        self.overrideAttr(urlutils, "MIN_ABS_FILEURL_LENGTH",
                          urlutils.WIN32_MIN_ABS_FILEURL_LENGTH)
        self.overrideAttr(osutils, "abspath", osutils._win32_abspath)
        self.overrideAttr(osutils, "normpath", osutils._win32_normpath)
        self.overrideAttr(osutils, "pathjoin", osutils._win32_pathjoin)
        self.overrideAttr(osutils, "split", osutils.ntpath.split)
        self.overrideAttr(osutils, "MIN_ABS_PATHLENGTH", 3)

    def test_same_url_posix(self):
        self._with_posix_paths()
        self.assertEqual("",
                         urlutils.file_relpath("file:///a", "file:///a"))
        self.assertEqual("",
                         urlutils.file_relpath("file:///a", "file:///a/"))
        self.assertEqual("",
                         urlutils.file_relpath("file:///a/", "file:///a"))

    def test_same_url_win32(self):
        self._with_win32_paths()
        self.assertEqual("",
                         urlutils.file_relpath("file:///A:/", "file:///A:/"))
        self.assertEqual("",
                         urlutils.file_relpath("file:///A|/", "file:///A:/"))
        self.assertEqual(
            "", urlutils.file_relpath("file:///A:/b/", "file:///A:/b/"))
        self.assertEqual(
            "", urlutils.file_relpath("file:///A:/b", "file:///A:/b/"))
        self.assertEqual(
            "", urlutils.file_relpath("file:///A:/b/", "file:///A:/b"))

    def test_child_posix(self):
        self._with_posix_paths()
        self.assertEqual(
            "b", urlutils.file_relpath("file:///a", "file:///a/b"))
        self.assertEqual(
            "b", urlutils.file_relpath("file:///a/", "file:///a/b"))
        self.assertEqual(
            "b/c", urlutils.file_relpath("file:///a", "file:///a/b/c"))

    def test_child_win32(self):
        self._with_win32_paths()
        self.assertEqual(
            "b", urlutils.file_relpath("file:///A:/", "file:///A:/b"))
        self.assertEqual(
            "b", urlutils.file_relpath("file:///A|/", "file:///A:/b"))
        self.assertEqual(
            "c", urlutils.file_relpath("file:///A:/b", "file:///A:/b/c"))
        self.assertEqual(
            "c", urlutils.file_relpath("file:///A:/b/", "file:///A:/b/c"))
        self.assertEqual(
            "c/d", urlutils.file_relpath("file:///A:/b", "file:///A:/b/c/d"))

    def test_sibling_posix(self):
        self._with_posix_paths()
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///a/b", "file:///a/c")
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///a/b/", "file:///a/c")
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///a/b/", "file:///a/c/")

    def test_sibling_win32(self):
        self._with_win32_paths()
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///A:/b", "file:///A:/c")
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///A:/b/", "file:///A:/c")
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///A:/b/", "file:///A:/c/")

    def test_parent_posix(self):
        self._with_posix_paths()
        self.assertRaises(PathNotChild,
                          urlutils.file_relpath, "file:///a/b", "file:///a")
        self.assertRaises(PathNotChild,
                          urlutils.file_relpath, "file:///a/b", "file:///a/")

    def test_parent_win32(self):
        self._with_win32_paths()
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///A:/b", "file:///A:/")
        self.assertRaises(
            PathNotChild,
            urlutils.file_relpath, "file:///A:/b/c", "file:///A:/b")


class QuoteTests(TestCase):

    def test_quote(self):
        self.assertEqual('abc%20def', urlutils.quote('abc def'))
        self.assertEqual('abc%2Fdef', urlutils.quote('abc/def', safe=''))
        self.assertEqual('abc/def', urlutils.quote('abc/def', safe='/'))

    def test_quote_tildes(self):
        # Whether ~ is quoted by default depends on the python version
        if sys.version_info[:2] >= (3, 7):
            # https://docs.python.org/3/whatsnew/3.7.html#urllib-parse
            self.assertEqual('~foo', urlutils.quote('~foo'))
        else:
            self.assertEqual('%7Efoo', urlutils.quote('~foo'))
        self.assertEqual('~foo', urlutils.quote('~foo', safe='/~'))

    def test_unquote(self):
        self.assertEqual('%', urlutils.unquote('%25'))
        self.assertEqual('\xe5', urlutils.unquote('%C3%A5'))
        self.assertEqual("\xe5", urlutils.unquote('\xe5'))

    def test_unquote_to_bytes(self):
        self.assertEqual(b'%', urlutils.unquote_to_bytes('%25'))
        self.assertEqual(b'\xc3\xa5', urlutils.unquote_to_bytes('%C3%A5'))



# Copyright (C) 2006 Canonical Ltd
# Copyright (C) 2008 Aaron Bentley <aaron@aaronbentley.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy.errors import BinaryFile
from breezy.iterablefile import IterableFile
from breezy.patch import (PatchInvokeError, diff3, iter_patched_from_hunks,
                          run_patch)
from breezy.patches import parse_patch
from breezy.tests import TestCase, TestCaseInTempDir


class TestPatch(TestCaseInTempDir):

    def test_diff3_binaries(self):
        with open('this', 'wb') as f:
            f.write(b'a')
        with open('other', 'wb') as f:
            f.write(b'a')
        with open('base', 'wb') as f:
            f.write(b'\x00')
        self.assertRaises(BinaryFile, diff3, 'unused', 'this', 'other', 'base')

    def test_missing_patch(self):
        self.assertRaises(PatchInvokeError, run_patch, '.', [],
                          _patch_cmd='/unlikely/to/exist')


class PatchesTester(TestCase):

    def datafile(self, filename):
        data_path = os.path.join(os.path.dirname(__file__),
                                 "test_patches_data", filename)
        return open(data_path, "rb")

    def data_lines(self, filename):
        with self.datafile(filename) as datafile:
            return datafile.readlines()

    def test_iter_patched_from_hunks(self):
        """Test a few patch files, and make sure they work."""
        files = [
            ('diff-2', 'orig-2', 'mod-2'),
            ('diff-3', 'orig-3', 'mod-3'),
            ('diff-4', 'orig-4', 'mod-4'),
            ('diff-5', 'orig-5', 'mod-5'),
            ('diff-6', 'orig-6', 'mod-6'),
            ('diff-7', 'orig-7', 'mod-7'),
        ]
        for diff, orig, mod in files:
            parsed = parse_patch(self.datafile(diff))
            orig_lines = list(self.datafile(orig))
            mod_lines = list(self.datafile(mod))
            iter_patched = iter_patched_from_hunks(orig_lines, parsed.hunks)
            patched_file = IterableFile(iter_patched)
            count = 0
            for patch_line in patched_file:
                self.assertEqual(patch_line, mod_lines[count], 'for file %s' % diff)
                count += 1
            self.assertEqual(count, len(mod_lines))



# Copyright (C) 2005-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import missing, tests
from ..missing import iter_log_revisions
from . import TestCaseWithTransport


class TestMissing(TestCaseWithTransport):

    def assertUnmerged(self, expected, source, target, restrict='all',
                       backward=False):
        unmerged = missing.find_unmerged(source, target, restrict=restrict,
                                         backward=backward)
        self.assertEqual(expected, unmerged)

    def test_find_unmerged(self):
        original_tree = self.make_branch_and_tree('original')
        original = original_tree.branch
        puller_tree = self.make_branch_and_tree('puller')
        puller = puller_tree.branch
        merger_tree = self.make_branch_and_tree('merger')
        merger = merger_tree.branch
        self.assertUnmerged(([], []), original, puller)
        original_tree.commit('a', rev_id=b'a')
        self.assertUnmerged(([('1', b'a', 0)], []), original, puller)
        puller_tree.pull(original)
        self.assertUnmerged(([], []), original, puller)
        merger_tree.pull(original)
        original_tree.commit('b', rev_id=b'b')
        original_tree.commit('c', rev_id=b'c')
        self.assertUnmerged(([('2', b'b', 0), ('3', b'c', 0)], []),
                            original, puller)
        self.assertUnmerged(([('3', b'c', 0), ('2', b'b', 0)], []),
                            original, puller, backward=True)

        puller_tree.pull(original)
        self.assertUnmerged(([], []), original, puller)
        self.assertUnmerged(([('2', b'b', 0), ('3', b'c', 0)], []),
                            original, merger)
        merger_tree.merge_from_branch(original)
        self.assertUnmerged(([('2', b'b', 0), ('3', b'c', 0)], []),
                            original, merger)
        merger_tree.commit('d', rev_id=b'd')
        self.assertUnmerged(([], [('2', b'd', 0)]), original, merger)

    def test_iter_log_revisions(self):
        base_tree = self.make_branch_and_tree('base')
        self.build_tree(['base/a'])
        base_tree.add(['a'], ids=[b'a-id'])
        base_tree.commit('add a', rev_id=b'b-1')

        child_tree = base_tree.controldir.sprout('child').open_workingtree()

        self.build_tree(['child/b'])
        child_tree.add(['b'], ids=[b'b-id'])
        child_tree.commit('adding b', rev_id=b'c-2')

        child_tree.remove(['a'])
        child_tree.commit('removing a', rev_id=b'c-3')

        self.build_tree_contents([('child/b', b'new contents for b\n')])
        child_tree.commit('modifying b', rev_id=b'c-4')

        child_tree.rename_one('b', 'c')
        child_tree.commit('rename b=>c', rev_id=b'c-5')

        base_extra, child_extra = missing.find_unmerged(base_tree.branch,
                                                        child_tree.branch)
        results = list(iter_log_revisions(base_extra,
                                          base_tree.branch.repository,
                                          verbose=True))
        self.assertEqual([], results)

        results = list(iter_log_revisions(child_extra,
                                          child_tree.branch.repository,
                                          verbose=True))
        self.assertEqual(4, len(results))

        r0, r1, r2, r3 = results

        self.assertEqual([('2', b'c-2'), ('3', b'c-3'),
                          ('4', b'c-4'), ('5', b'c-5'), ],
                         [(r.revno, r.rev.revision_id) for r in results])

        delta0 = r0.delta
        self.assertNotEqual(None, delta0)
        self.assertEqual([('b', 'file')], [(c.path[1], c.kind[1]) for c in delta0.added])
        self.assertEqual([], delta0.removed)
        self.assertEqual([], delta0.renamed)
        self.assertEqual([], delta0.copied)
        self.assertEqual([], delta0.modified)

        delta1 = r1.delta
        self.assertNotEqual(None, delta1)
        self.assertEqual([], delta1.added)
        self.assertEqual([('a', 'file')], [(c.path[0], c.kind[0]) for c in delta1.removed])
        self.assertEqual([], delta1.renamed)
        self.assertEqual([], delta1.copied)
        self.assertEqual([], delta1.modified)

        delta2 = r2.delta
        self.assertNotEqual(None, delta2)
        self.assertEqual([], delta2.added)
        self.assertEqual([], delta2.removed)
        self.assertEqual([], delta2.renamed)
        self.assertEqual([], delta2.copied)
        self.assertEqual(
            [('b', 'file', True, False)],
            [(c.path[1], c.kind[1], c.changed_content, c.meta_modified()) for c in delta2.modified])

        delta3 = r3.delta
        self.assertNotEqual(None, delta3)
        self.assertEqual([], delta3.added)
        self.assertEqual([], delta3.copied)
        self.assertEqual([], delta3.removed)
        self.assertEqual(
            [('b', 'c', 'file', False, False)],
            [(c.path[0], c.path[1], c.kind[1], c.changed_content, c.meta_modified())
                for c in delta3.renamed])
        self.assertEqual([], delta3.modified)


class TestFindUnmerged(tests.TestCaseWithTransport):

    def assertUnmerged(self, local, remote, local_branch, remote_branch,
                       restrict='all', include_merged=False, backward=False,
                       local_revid_range=None, remote_revid_range=None):
        """Check the output of find_unmerged_mainline_revisions"""
        local_extra, remote_extra = missing.find_unmerged(
            local_branch, remote_branch, restrict,
            include_merged=include_merged, backward=backward,
            local_revid_range=local_revid_range,
            remote_revid_range=remote_revid_range)
        self.assertEqual(local, local_extra)
        self.assertEqual(remote, remote_extra)

    def test_same_branch(self):
        tree = self.make_branch_and_tree('tree')
        rev1 = tree.commit('one')
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertUnmerged([], [], tree.branch, tree.branch)
        self.assertUnmerged([], [], tree.branch, tree.branch,
                            local_revid_range=(rev1, rev1))

    def test_one_ahead(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('one')
        tree2 = tree.controldir.sprout('tree2').open_workingtree()
        rev2 = tree2.commit('two')
        self.assertUnmerged([], [('2', rev2, 0)], tree.branch, tree2.branch)
        self.assertUnmerged([('2', rev2, 0)], [], tree2.branch, tree.branch)

    def test_restrict(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('one')
        tree2 = tree.controldir.sprout('tree2').open_workingtree()
        rev2 = tree2.commit('two')
        self.assertUnmerged([], [('2', rev2, 0)], tree.branch, tree2.branch)
        self.assertUnmerged([], None, tree.branch, tree2.branch, 'local')
        self.assertUnmerged(None, [('2', rev2, 0)], tree.branch, tree2.branch,
                            'remote')

    def test_merged(self):
        tree = self.make_branch_and_tree('tree')
        rev1 = tree.commit('one')
        tree2 = tree.controldir.sprout('tree2').open_workingtree()
        tree2.commit('two')
        tree2.commit('three')
        tree.merge_from_branch(tree2.branch)
        rev4 = tree.commit('four')

        self.assertUnmerged([('2', rev4, 0)], [], tree.branch, tree2.branch)
        self.assertUnmerged([('2', rev4, 0)], [], tree.branch, tree2.branch,
                            local_revid_range=(rev4, rev4))
        self.assertUnmerged([], [], tree.branch, tree2.branch,
                            local_revid_range=(rev1, rev1))

    def test_include_merged(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('one', rev_id=b'rev1')

        tree2 = tree.controldir.sprout('tree2').open_workingtree()
        tree2.commit('two', rev_id=b'rev2')
        rev3 = tree2.commit('three', rev_id=b'rev3')

        tree3 = tree2.controldir.sprout('tree3').open_workingtree()
        rev4 = tree3.commit('four', rev_id=b'rev4')
        rev5 = tree3.commit('five', rev_id=b'rev5')

        tree2.merge_from_branch(tree3.branch)
        rev6 = tree2.commit('six', rev_id=b'rev6')

        self.assertUnmerged([], [('2', b'rev2', 0), ('3', b'rev3', 0),
                                 ('4', b'rev6', 0),
                                 ('3.1.1', b'rev4', 1), ('3.1.2', b'rev5', 1),
                                 ],
                            tree.branch, tree2.branch,
                            include_merged=True)

        self.assertUnmerged([], [('4', b'rev6', 0),
                                 ('3.1.2', b'rev5', 1), ('3.1.1', b'rev4', 1),
                                 ('3', b'rev3', 0), ('2', b'rev2', 0),
                                 ],
                            tree.branch, tree2.branch,
                            include_merged=True,
                            backward=True)

        self.assertUnmerged(
            [], [('4', b'rev6', 0)],
            tree.branch, tree2.branch,
            include_merged=True, remote_revid_range=(rev6, rev6))

        self.assertUnmerged(
            [], [('3', b'rev3', 0), ('3.1.1', b'rev4', 1)],
            tree.branch, tree2.branch,
            include_merged=True, remote_revid_range=(rev3, rev4))

        self.assertUnmerged(
            [], [('4', b'rev6', 0), ('3.1.2', b'rev5', 1)],
            tree.branch, tree2.branch,
            include_merged=True, remote_revid_range=(rev5, rev6))

    def test_revision_range(self):
        local = self.make_branch_and_tree('local')
        lrevid1 = local.commit('one')
        remote = local.controldir.sprout('remote').open_workingtree()
        rrevid2 = remote.commit('two')
        rrevid3 = remote.commit('three')
        rrevid4 = remote.commit('four')
        lrevid2 = local.commit('two')
        lrevid3 = local.commit('three')
        lrevid4 = local.commit('four')
        local_extra = [('2', lrevid2, 0), ('3', lrevid3, 0), ('4', lrevid4, 0)]
        remote_extra = [('2', rrevid2, 0), ('3', rrevid3, 0),
                        ('4', rrevid4, 0)]

        # control
        self.assertUnmerged(local_extra, remote_extra,
                            local.branch, remote.branch)
        self.assertUnmerged(local_extra, remote_extra,
                            local.branch, remote.branch, local_revid_range=(
                                None, None),
                            remote_revid_range=(None, None))

        # exclude local revisions
        self.assertUnmerged(
            [('2', lrevid2, 0)], remote_extra,
            local.branch, remote.branch, local_revid_range=(lrevid2, lrevid2))
        self.assertUnmerged(
            [('2', lrevid2, 0), ('3', lrevid3, 0)], remote_extra,
            local.branch, remote.branch, local_revid_range=(lrevid2, lrevid3))
        self.assertUnmerged(
            [('2', lrevid2, 0), ('3', lrevid3, 0)], None,
            local.branch, remote.branch, 'local',
            local_revid_range=(lrevid2, lrevid3))

        # exclude remote revisions
        self.assertUnmerged(
            local_extra, [('2', rrevid2, 0)],
            local.branch, remote.branch, remote_revid_range=(None, rrevid2))
        self.assertUnmerged(
            local_extra, [('2', rrevid2, 0)],
            local.branch, remote.branch, remote_revid_range=(lrevid1, rrevid2))
        self.assertUnmerged(
            local_extra, [('2', rrevid2, 0)],
            local.branch, remote.branch, remote_revid_range=(rrevid2, rrevid2))
        self.assertUnmerged(
            local_extra, [('2', rrevid2, 0), ('3', rrevid3, 0)],
            local.branch, remote.branch, remote_revid_range=(None, rrevid3))
        self.assertUnmerged(
            local_extra, [('2', rrevid2, 0), ('3', rrevid3, 0)],
            local.branch, remote.branch, remote_revid_range=(rrevid2, rrevid3))
        self.assertUnmerged(
            local_extra, [('3', rrevid3, 0)],
            local.branch, remote.branch, remote_revid_range=(rrevid3, rrevid3))
        self.assertUnmerged(None, [('2', rrevid2, 0), ('3', rrevid3, 0)],
                            local.branch, remote.branch, 'remote',
                            remote_revid_range=(rrevid2, rrevid3))

        # exclude local and remote revisions
        self.assertUnmerged([('3', lrevid3, 0)], [('3', rrevid3, 0)],
                            local.branch, remote.branch, local_revid_range=(
                                lrevid3, lrevid3),
                            remote_revid_range=(rrevid3, rrevid3))



# Copyright (C) 2007, 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Counted lock class"""

from . import errors


class CountedLock:
    """Decorator around a lock that makes it reentrant.

    This can be used with any object that provides a basic Lock interface,
    including LockDirs and OS file locks.

    :ivar _token: While a write lock is held, this is the token
        for it.
    """

    def __init__(self, real_lock):
        self._real_lock = real_lock
        self._lock_mode = None
        self._lock_count = 0

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__,
                           self._real_lock)

    def break_lock(self):
        self._real_lock.break_lock()
        self._lock_mode = None
        self._lock_count = 0

    def get_physical_lock_status(self):
        """Return physical lock status.

        Returns true if a lock is held on the transport. If no lock is held, or
        the underlying locking mechanism does not support querying lock
        status, false is returned.
        """
        try:
            return self._real_lock.peek() is not None
        except NotImplementedError:
            return False

    def is_locked(self):
        return self._lock_mode is not None

    def lock_read(self):
        """Acquire the lock in read mode.

        If the lock is already held in either read or write mode this
        increments the count and succeeds.  If the lock is not already held,
        it is taken in read mode.
        """
        if self._lock_mode:
            self._lock_count += 1
        else:
            self._real_lock.lock_read()
            self._lock_count = 1
            self._lock_mode = 'r'

    def lock_write(self, token=None):
        """Acquire the lock in write mode.

        If the lock was originally acquired in read mode this will fail.

        :param token: If given and the lock is already held,
            then validate that we already hold the real
            lock with this token.

        :returns: The token from the underlying lock.
        """
        if self._lock_count == 0:
            self._token = self._real_lock.lock_write(token=token)
            self._lock_mode = 'w'
            self._lock_count += 1
            return self._token
        elif self._lock_mode != 'w':
            raise errors.ReadOnlyError(self)
        else:
            self._real_lock.validate_token(token)
            self._lock_count += 1
            return self._token

    def unlock(self):
        if self._lock_count == 0:
            raise errors.LockNotHeld(self)
        elif self._lock_count == 1:
            # these are decremented first; if we fail to unlock the most
            # reasonable assumption is that we still don't have the lock
            # anymore
            self._lock_mode = None
            self._lock_count -= 1
            self._real_lock.unlock()
        else:
            self._lock_count -= 1



# Copyright (C) 2005, 2006, 2008, 2009, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for revision properties."""

from breezy.tests import TestNotApplicable
from breezy.tests.per_repository import TestCaseWithRepository


class TestRevProps(TestCaseWithRepository):

    def test_simple_revprops(self):
        """Simple revision properties"""
        wt = self.make_branch_and_tree('.')
        b = wt.branch
        b.nick = 'Nicholas'
        if b.repository._format.supports_custom_revision_properties:
            props = {'flavor': 'choc-mint',
                     'condiment': 'orange\n  mint\n\tcandy',
                     'empty': '',
                     'non_ascii': '\xb5'}
        else:
            props = {}
        rev1 = wt.commit(message='initial null commit',
                         revprops=props,
                         allow_pointless=True)
        rev = b.repository.get_revision(rev1)
        if b.repository._format.supports_custom_revision_properties:
            self.assertTrue('flavor' in rev.properties)
            self.assertEqual(rev.properties['flavor'], 'choc-mint')
            expected_revprops = {
                'condiment': 'orange\n  mint\n\tcandy',
                'empty': '',
                'flavor': 'choc-mint',
                'non_ascii': '\xb5',
                }
        else:
            expected_revprops = {}
        if b.repository._format.supports_storing_branch_nick:
            expected_revprops['branch-nick'] = 'Nicholas'
        for name, value in expected_revprops.items():
            self.assertEqual(rev.properties[name], value)

    def test_invalid_revprops(self):
        """Invalid revision properties"""
        wt = self.make_branch_and_tree('.')
        b = wt.branch
        if not b.repository._format.supports_custom_revision_properties:
            raise TestNotApplicable(
                'format does not support custom revision properties')
        self.assertRaises(ValueError,
                          wt.commit,
                          message='invalid',
                          revprops={'what a silly property': 'fine'})
        self.assertRaises(ValueError,
                          wt.commit,
                          message='invalid',
                          revprops=dict(number=13))


class TestRevisionAttributes(TestCaseWithRepository):
    """Test that revision attributes are correct."""

    def test_revision_accessors(self):
        """Make sure the values that come out of a revision are the
        same as the ones that go in.
        """
        tree1 = self.make_branch_and_tree("br1")
        if tree1.branch.repository._format.supports_custom_revision_properties:
            revprops = {'empty': '',
                        'value': 'one',
                        'unicode': '\xb5',
                        'multiline': 'foo\nbar\n\n'
                        }
        else:
            revprops = {}
        # create a revision
        rev1 = tree1.commit(message="quux", allow_pointless=True,
                            committer="jaq", revprops=revprops)
        self.assertEqual(tree1.branch.last_revision(), rev1)
        rev_a = tree1.branch.repository.get_revision(
            tree1.branch.last_revision())

        tree2 = self.make_branch_and_tree("br2")
        tree2.commit(message=rev_a.message,
                     timestamp=rev_a.timestamp,
                     timezone=rev_a.timezone,
                     committer=rev_a.committer,
                     rev_id=(rev_a.revision_id
                             if tree2.branch.repository._format.supports_setting_revision_ids
                             else None),
                     revprops=rev_a.properties,
                     allow_pointless=True,  # there's nothing in this commit
                     strict=True,
                     verbose=True)
        rev_b = tree2.branch.repository.get_revision(
            tree2.branch.last_revision())

        self.assertEqual(rev_a.message, rev_b.message)
        self.assertEqual(rev_a.timestamp, rev_b.timestamp)
        self.assertEqual(rev_a.timezone, rev_b.timezone)
        self.assertEqual(rev_a.committer, rev_b.committer)
        self.assertEqual(rev_a.revision_id, rev_b.revision_id)
        self.assertEqual(rev_a.properties, rev_b.properties)

    def test_zero_timezone(self):
        tree1 = self.make_branch_and_tree("br1")

        # create a revision
        r1 = tree1.commit(message="quux", timezone=0)
        rev_a = tree1.branch.repository.get_revision(r1)
        self.assertEqual(0, rev_a.timezone)



# Copyright (C) 2007-2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ...builtins import cmd_init_shared_repository
from .. import transport_util, ui_testing


class TestInitRepository(transport_util.TestCaseWithConnectionHookedTransport):

    def setUp(self):
        super().setUp()
        self.start_logging_connections()

    def test_init_shared_repository(self):
        cmd = cmd_init_shared_repository()
        # We don't care about the ouput but 'outf' should be defined
        cmd.outf = ui_testing.StringIOWithEncoding()
        cmd.run(self.get_url())
        self.assertEqual(1, len(self.connections))



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Branch.dotted_revno_to_revision_id()"""

from breezy import errors
from breezy.bzr.fullhistory import FullHistoryBzrBranch
from breezy.tests import TestNotApplicable
from breezy.tests.per_branch import TestCaseWithBranch


class TestDottedRevnoToRevisionId(TestCaseWithBranch):

    def test_lookup_revision_id_by_dotted(self):
        tree, revmap = self.create_tree_with_merge()
        the_branch = tree.branch
        the_branch.lock_read()
        self.addCleanup(the_branch.unlock)
        self.assertEqual(
            b'null:', the_branch.dotted_revno_to_revision_id((0,)))
        self.assertEqual(revmap['1'],
                         the_branch.dotted_revno_to_revision_id((1, )))
        self.assertEqual(revmap['2'],
                         the_branch.dotted_revno_to_revision_id((2, )))
        self.assertEqual(revmap['3'],
                         the_branch.dotted_revno_to_revision_id((3, )))
        self.assertEqual(revmap['1.1.1'],
                         the_branch.dotted_revno_to_revision_id(
            (1, 1, 1)))
        self.assertRaises(
            errors.NoSuchRevision,
            the_branch.dotted_revno_to_revision_id, (1, 0, 2))
        # Test reverse caching
        self.assertEqual(
            None,
            the_branch._partial_revision_id_to_revno_cache.get(revmap['1']))
        self.assertEqual(
            revmap['1'], the_branch.dotted_revno_to_revision_id(
                (1, ), _cache_reverse=True))
        self.assertEqual(
            (1,),
            the_branch._partial_revision_id_to_revno_cache.get(revmap['1']))

    def test_ghost_revision(self):
        if not self.bzrdir_format.repository_format.supports_ghosts:
            raise TestNotApplicable("repository format does not support ghosts")
        builder = self.make_branch_builder('foo')
        builder.start_series()
        try:
            revid1 = builder.build_snapshot(
                [b'ghost'], [('add', ('', b'ROOT_ID', 'directory', ''))],
                allow_leftmost_as_ghost=True, revision_id=b'tip')
        finally:
            builder.finish_series()
        b = builder.get_branch()
        if isinstance(b, FullHistoryBzrBranch):
            raise TestNotApplicable("branch format stores full history")
        b.set_last_revision_info(4, revid1)
        self.assertRaises(errors.GhostRevisionsHaveNoRevno, b.dotted_revno_to_revision_id, (2,))



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Code to estimate the entropy of content"""

import zlib


class ZLibEstimator:
    """Uses zlib.compressobj to estimate compressed size."""

    def __init__(self, target_size, min_compression=2.0):
        """Create a new estimator.

        :param target_size: The desired size of the compressed content.
        :param min_compression: Estimated minimum compression. By default we
            assume that the content is 'text', which means a min compression of
            about 2:1.
        """
        self._target_size = target_size
        self._compressor = zlib.compressobj()
        self._uncompressed_size_added = 0
        self._compressed_size_added = 0
        self._unflushed_size_added = 0
        self._estimated_compression = 2.0

    def add_content(self, content):
        self._uncompressed_size_added += len(content)
        self._unflushed_size_added += len(content)
        z_size = len(self._compressor.compress(content))
        if z_size > 0:
            self._record_z_len(z_size)

    def _record_z_len(self, count):
        # We got some compressed bytes, update the counters
        self._compressed_size_added += count
        self._unflushed_size_added = 0
        # So far we've read X uncompressed bytes, and written Y compressed
        # bytes. We should have a decent estimate of the final compression.
        self._estimated_compression = (float(self._uncompressed_size_added)
                                       / self._compressed_size_added)

    def full(self):
        """Have we reached the target size?"""
        if self._unflushed_size_added:
            remaining_size = self._target_size - self._compressed_size_added
            # Estimate how much compressed content the unflushed data will
            # consume
            est_z_size = (self._unflushed_size_added /
                          self._estimated_compression)
            if est_z_size >= remaining_size:
                # We estimate we are close to remaining
                z_size = len(self._compressor.flush(zlib.Z_SYNC_FLUSH))
                self._record_z_len(z_size)
        return self._compressed_size_added >= self._target_size



# Copyright (C) 2005 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for the revision/inventory Serializers."""


from .. import chk_serializer, serializer, xml5, xml6, xml7, xml8
from . import TestCase


class TestSerializer(TestCase):
    """Test serializer"""

    def test_registry(self):
        self.assertIs(xml5.serializer_v5,
                      serializer.format_registry.get('5'))
        self.assertIs(xml6.serializer_v6,
                      serializer.format_registry.get('6'))
        self.assertIs(xml7.serializer_v7,
                      serializer.format_registry.get('7'))
        self.assertIs(xml8.serializer_v8,
                      serializer.format_registry.get('8'))
        self.assertIs(chk_serializer.chk_serializer_255_bigpage,
                      serializer.format_registry.get('9'))



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functionality for doing annotations in the 'optimal' way"""

from . import errors
from . import graph as _mod_graph
from . import osutils, ui


class Annotator:
    """Class that drives performing annotations."""

    def __init__(self, vf):
        """Create a new Annotator from a VersionedFile."""
        self._vf = vf
        self._parent_map = {}
        self._text_cache = {}
        # Map from key => number of nexts that will be built from this key
        self._num_needed_children = {}
        self._annotations_cache = {}
        self._heads_provider = None
        self._ann_tuple_cache = {}

    def _update_needed_children(self, key, parent_keys):
        for parent_key in parent_keys:
            if parent_key in self._num_needed_children:
                self._num_needed_children[parent_key] += 1
            else:
                self._num_needed_children[parent_key] = 1

    def _get_needed_keys(self, key):
        """Determine the texts we need to get from the backing vf.

        :return: (vf_keys_needed, ann_keys_needed)
            vf_keys_needed  These are keys that we need to get from the vf
            ann_keys_needed Texts which we have in self._text_cache but we
                            don't have annotations for. We need to yield these
                            in the proper order so that we can get proper
                            annotations.
        """
        parent_map = self._parent_map
        # We need 1 extra copy of the node we will be looking at when we are
        # done
        self._num_needed_children[key] = 1
        vf_keys_needed = set()
        ann_keys_needed = set()
        needed_keys = {key}
        while needed_keys:
            parent_lookup = []
            next_parent_map = {}
            for key in needed_keys:
                if key in self._parent_map:
                    # We don't need to lookup this key in the vf
                    if key not in self._text_cache:
                        # Extract this text from the vf
                        vf_keys_needed.add(key)
                    elif key not in self._annotations_cache:
                        # We do need to annotate
                        ann_keys_needed.add(key)
                        next_parent_map[key] = self._parent_map[key]
                else:
                    parent_lookup.append(key)
                    vf_keys_needed.add(key)
            needed_keys = set()
            next_parent_map.update(self._vf.get_parent_map(parent_lookup))
            for key, parent_keys in next_parent_map.items():
                if parent_keys is None:  # No graph versionedfile
                    parent_keys = ()
                    next_parent_map[key] = ()
                self._update_needed_children(key, parent_keys)
                needed_keys.update([key for key in parent_keys
                                    if key not in parent_map])
            parent_map.update(next_parent_map)
            # _heads_provider does some graph caching, so it is only valid
            # while self._parent_map hasn't changed
            self._heads_provider = None
        return vf_keys_needed, ann_keys_needed

    def _get_needed_texts(self, key, pb=None):
        """Get the texts we need to properly annotate key.

        :param key: A Key that is present in self._vf
        :return: Yield (this_key, text, num_lines)
            'text' is an opaque object that just has to work with whatever
            matcher object we are using. Currently it is always 'lines' but
            future improvements may change this to a simple text string.
        """
        keys, ann_keys = self._get_needed_keys(key)
        if pb is not None:
            pb.update('getting stream', 0, len(keys))
        stream = self._vf.get_record_stream(keys, 'topological', True)
        for idx, record in enumerate(stream):
            if pb is not None:
                pb.update('extracting', 0, len(keys))
            if record.storage_kind == 'absent':
                raise errors.RevisionNotPresent(record.key, self._vf)
            this_key = record.key
            lines = record.get_bytes_as('lines')
            num_lines = len(lines)
            self._text_cache[this_key] = lines
            yield this_key, lines, num_lines
        for key in ann_keys:
            lines = self._text_cache[key]
            num_lines = len(lines)
            yield key, lines, num_lines

    def _get_parent_annotations_and_matches(self, key, text, parent_key):
        """Get the list of annotations for the parent, and the matching lines.

        :param text: The opaque value given by _get_needed_texts
        :param parent_key: The key for the parent text
        :return: (parent_annotations, matching_blocks)
            parent_annotations is a list as long as the number of lines in
                parent
            matching_blocks is a list of (parent_idx, text_idx, len) tuples
                indicating which lines match between the two texts
        """
        parent_lines = self._text_cache[parent_key]
        parent_annotations = self._annotations_cache[parent_key]
        # PatienceSequenceMatcher should probably be part of Policy
        from patiencediff import PatienceSequenceMatcher
        matcher = PatienceSequenceMatcher(
            None, parent_lines, text)
        matching_blocks = matcher.get_matching_blocks()
        return parent_annotations, matching_blocks

    def _update_from_first_parent(self, key, annotations, lines, parent_key):
        """Reannotate this text relative to its first parent."""
        (parent_annotations,
         matching_blocks) = self._get_parent_annotations_and_matches(
             key, lines, parent_key)

        for parent_idx, lines_idx, match_len in matching_blocks:
            # For all matching regions we copy across the parent annotations
            annotations[lines_idx:lines_idx + match_len] = \
                parent_annotations[parent_idx:parent_idx + match_len]

    def _update_from_other_parents(self, key, annotations, lines,
                                   this_annotation, parent_key):
        """Reannotate this text relative to a second (or more) parent."""
        (parent_annotations,
         matching_blocks) = self._get_parent_annotations_and_matches(
             key, lines, parent_key)

        last_ann = None
        last_parent = None
        last_res = None
        # TODO: consider making all annotations unique and then using 'is'
        #       everywhere. Current results claim that isn't any faster,
        #       because of the time spent deduping
        #       deduping also saves a bit of memory. For NEWS it saves ~1MB,
        #       but that is out of 200-300MB for extracting everything, so a
        #       fairly trivial amount
        for parent_idx, lines_idx, match_len in matching_blocks:
            # For lines which match this parent, we will now resolve whether
            # this parent wins over the current annotation
            ann_sub = annotations[lines_idx:lines_idx + match_len]
            par_sub = parent_annotations[parent_idx:parent_idx + match_len]
            if ann_sub == par_sub:
                continue
            for idx in range(match_len):
                ann = ann_sub[idx]
                par_ann = par_sub[idx]
                ann_idx = lines_idx + idx
                if ann == par_ann:
                    # Nothing to change
                    continue
                if ann == this_annotation:
                    # Originally claimed 'this', but it was really in this
                    # parent
                    annotations[ann_idx] = par_ann
                    continue
                # Resolve the fact that both sides have a different value for
                # last modified
                if ann == last_ann and par_ann == last_parent:
                    annotations[ann_idx] = last_res
                else:
                    new_ann = set(ann)
                    new_ann.update(par_ann)
                    new_ann = tuple(sorted(new_ann))
                    annotations[ann_idx] = new_ann
                    last_ann = ann
                    last_parent = par_ann
                    last_res = new_ann

    def _record_annotation(self, key, parent_keys, annotations):
        self._annotations_cache[key] = annotations
        for parent_key in parent_keys:
            num = self._num_needed_children[parent_key]
            num -= 1
            if num == 0:
                del self._text_cache[parent_key]
                del self._annotations_cache[parent_key]
                # Do we want to clean up _num_needed_children at this point as
                # well?
            self._num_needed_children[parent_key] = num

    def _annotate_one(self, key, text, num_lines):
        this_annotation = (key,)
        # Note: annotations will be mutated by calls to _update_from*
        annotations = [this_annotation] * num_lines
        parent_keys = self._parent_map[key]
        if parent_keys:
            self._update_from_first_parent(key, annotations, text,
                                           parent_keys[0])
            for parent in parent_keys[1:]:
                self._update_from_other_parents(key, annotations, text,
                                                this_annotation, parent)
        self._record_annotation(key, parent_keys, annotations)

    def add_special_text(self, key, parent_keys, text):
        """Add a specific text to the graph.

        This is used to add a text which is not otherwise present in the
        versioned file. (eg. a WorkingTree injecting 'current:' into the
        graph to annotate the edited content.)

        :param key: The key to use to request this text be annotated
        :param parent_keys: The parents of this text
        :param text: A string containing the content of the text
        """
        self._parent_map[key] = parent_keys
        self._text_cache[key] = osutils.split_lines(text)
        self._heads_provider = None

    def annotate(self, key):
        """Return annotated fulltext for the given key.

        :param key: A tuple defining the text to annotate
        :return: ([annotations], [lines])
            annotations is a list of tuples of keys, one for each line in lines
                        each key is a possible source for the given line.
            lines the text of "key" as a list of lines
        """
        with ui.ui_factory.nested_progress_bar() as pb:
            for text_key, text, num_lines in self._get_needed_texts(
                    key, pb=pb):
                self._annotate_one(text_key, text, num_lines)
        try:
            annotations = self._annotations_cache[key]
        except KeyError as exc:
            raise errors.RevisionNotPresent(key, self._vf) from exc
        return annotations, self._text_cache[key]

    def _get_heads_provider(self):
        if self._heads_provider is None:
            self._heads_provider = _mod_graph.KnownGraph(self._parent_map)
        return self._heads_provider

    def _resolve_annotation_tie(self, the_heads, line, tiebreaker):
        if tiebreaker is None:
            head = sorted(the_heads)[0]
        else:
            # Backwards compatibility, break up the heads into pairs and
            # resolve the result
            next_head = iter(the_heads)
            head = next(next_head)
            for possible_head in next_head:
                annotated_lines = ((head, line), (possible_head, line))
                head = tiebreaker(annotated_lines)[0]
        return head

    def annotate_flat(self, key):
        """Determine the single-best-revision to source for each line.

        This is meant as a compatibility thunk to how annotate() used to work.
        :return: [(ann_key, line)]
            A list of tuples with a single annotation key for each line.
        """
        from .annotate import _break_annotation_tie
        custom_tiebreaker = _break_annotation_tie
        annotations, lines = self.annotate(key)
        out = []
        heads = self._get_heads_provider().heads
        append = out.append
        for annotation, line in zip(annotations, lines):
            if len(annotation) == 1:
                head = annotation[0]
            else:
                the_heads = heads(annotation)
                if len(the_heads) == 1:
                    for head in the_heads:
                        break  # get the item out of the set
                else:
                    head = self._resolve_annotation_tie(the_heads, line,
                                                        custom_tiebreaker)
            append((head, line))
        return out



# Copyright (C) 2007, 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import random
import time

from breezy import tests, timestamp
from breezy.osutils import local_time_offset


class TestPatchHeader(tests.TestCase):

    def test_format_patch_date(self):
        # epoch is always in utc
        self.assertEqual('1970-01-01 00:00:00 +0000',
                         timestamp.format_patch_date(0))
        self.assertEqual('1970-01-01 00:00:00 +0000',
                         timestamp.format_patch_date(0, 5 * 3600))
        self.assertEqual('1970-01-01 00:00:00 +0000',
                         timestamp.format_patch_date(0, -5 * 3600))
        # regular timestamp with typical timezone
        self.assertEqual('2007-03-06 10:04:19 -0500',
                         timestamp.format_patch_date(1173193459, -5 * 3600))
        # the timezone part is HHMM
        self.assertEqual('2007-03-06 09:34:19 -0530',
                         timestamp.format_patch_date(1173193459, -5.5 * 3600))
        # timezones can be offset by single minutes (but no less)
        self.assertEqual('2007-03-06 15:05:19 +0001',
                         timestamp.format_patch_date(1173193459, +1 * 60))

    def test_parse_patch_date(self):
        self.assertEqual(
            (0, 0),
            timestamp.parse_patch_date('1970-01-01 00:00:00 +0000'))
        # even though we don't emit pre-epoch dates, we can parse them
        self.assertEqual(
            (0, -5 * 3600),
            timestamp.parse_patch_date('1969-12-31 19:00:00 -0500'))
        self.assertEqual(
            (0, +5 * 3600),
            timestamp.parse_patch_date('1970-01-01 05:00:00 +0500'))
        self.assertEqual(
            (1173193459, -5 * 3600),
            timestamp.parse_patch_date('2007-03-06 10:04:19 -0500'))
        # offset of three minutes
        self.assertEqual(
            (1173193459, +3 * 60),
            timestamp.parse_patch_date('2007-03-06 15:07:19 +0003'))
        # No space between time and offset
        self.assertEqual(
            (1173193459, -5 * 3600),
            timestamp.parse_patch_date('2007-03-06 10:04:19-0500'))
        # Extra spacing
        self.assertEqual(
            (1173193459, -5 * 3600),
            timestamp.parse_patch_date('2007-03-06     10:04:19     -0500'))

    def test_parse_patch_date_bad(self):
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          'NOT A TIME')
        # Extra data at end
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 -0500x')
        # Missing day
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03 10:04:19 -0500')
        # Missing seconds
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04 -0500')
        # Missing offset
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19')
        # Missing plus or minus in offset
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 0500')
        # Invalid hour in offset
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 +2400')
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 -2400')
        # Invalid minute in offset
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 -0560')
        # Too many digits in offset
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 79500')
        # Minus sign in middle of offset
        self.assertRaises(ValueError, timestamp.parse_patch_date,
                          '2007-03-06 10:04:19 +05-5')


class UnpackHighresDateTests(tests.TestCase):

    def test_unpack_highres_date(self):
        self.assertEqual(
            (1120153132.3508501, -18000),
            timestamp.unpack_highres_date(
                'Thu 2005-06-30 12:38:52.350850105 -0500'))
        self.assertEqual(
            (1120153132.3508501, 0),
            timestamp.unpack_highres_date(
                'Thu 2005-06-30 17:38:52.350850105 +0000'))
        self.assertEqual(
            (1120153132.3508501, 7200),
            timestamp.unpack_highres_date(
                'Thu 2005-06-30 19:38:52.350850105 +0200'))
        self.assertEqual(
            (1152428738.867522, 19800),
            timestamp.unpack_highres_date(
                'Sun 2006-07-09 12:35:38.867522001 +0530'))

    def test_random(self):
        t = time.time()
        o = local_time_offset()
        t2, o2 = timestamp.unpack_highres_date(
            timestamp.format_highres_date(t, o))
        self.assertEqual(t, t2)
        self.assertEqual(o, o2)
        t -= 24 * 3600 * 365 * 2  # Start 2 years ago
        o = -12 * 3600
        for count in range(500):
            t += random.random() * 24 * 3600 * 30
            try:
                time.gmtime(t + o)
            except (OverflowError, ValueError):
                # We've reached the maximum for time_t on this platform
                break
            if time.localtime(t).tm_year > 9998:
                # strptime doesn't always understand years with more than 4
                # digits.
                break
            # Add 1 wrap around from [-12, 12]
            o = ((o / 3600 + 13) % 25 - 12) * 3600
            date = timestamp.format_highres_date(t, o)
            t2, o2 = timestamp.unpack_highres_date(date)
            self.assertEqual(
                t, t2,
                'Failed on date {!r}, {},{} diff:{}'.format(date, t, o, t2 - t))
            self.assertEqual(
                o, o2,
                'Failed on date {!r}, {},{} diff:{}'.format(date, t, o, t2 - t))
